# This file is part of pyZohoAPI, Copyright (C) Todd D. Esposito 2021.
# Distributed under the MIT License (see https://opensource.org/licenses/MIT).

import datetime
import logging
from time import sleep

import requests

from .collection import DottedDict, DottedList
from .utils import diff
from ..exceptions import *

logging.getLogger('pyzohoapi').addHandler(logging.NullHandler())

class ZohoAPIBase:
    _regionmap = {
        'australia': "com.au",
        'com.au': "com.au",
        'com': "com",
        'eu': "eu",
        'europe': "eu",
        'in': "in",
        'india': "in",
        'us': "com",
    }
    def __init__(self, organization_id, region="us", **apiArgs):
        """ Constructor

        :param organization_id: Zoho Organization ID to which to connect
        :type organization_id: str
        :param region: Zoho Data Center Region. Defaults to "us".
        :type region: str
        :param **apiArgs: additional parameters for API operation.
        :raises ZohoUnknownRegionException: if the region is unknown or invalid.

        """
        region = region.lower()
        if region not in self._regionmap:
            raise ZohoUnknownRegionException(region)
        self._org = organization_id
        self._endpoint = self.get_endpoint(region)
        self._oauth = f"https://accounts.zoho.{self._regionmap[region]}/oauth/v2"
        self._ratelimit = {
            'limit': None,
            'NextCall': datetime.datetime.now().timestamp(),
            'remaining': 99999999,
            'reset': None,
        }
        self._api_keys = {
            'access_token': None,
            'client_id': None,
            'client_secret': None,
            'intercall_delay': 0,
            'max_retries': 10,
            'max_retry_after': 180,
            'min_calls_remaining': 1,
            'redirect_url': None,
            'refresh_token': None,
            'retry_backoff_seconds': 0.5,
        }
        self.update_tokens(apiArgs)
        self._logger = logging.getLogger('pyzohoapi')

    def auth_header(self):
        """ Returns the authorization header, refreshing the access_token as needed.

        :return: {'Authorization': '... access token ...'}
        :raises ZohoAuthRefreshFailure: if a refresh attemp fails.
        :raises ZohoInsufficentAuthKeys: if we don't have enough info to refresh.
        """
        if self._api_keys.get('access_token') and self._api_keys['AccessExpiresAt'] > datetime.datetime.now().timestamp():
            return {'Authorization': f"Zoho-oauthtoken {self._api_keys['access_token']}"}
        if self._api_keys.get('refresh_token'):
            self.log("requesting new access token")
            rsp = requests.post(f"{self._oauth}/token", params={
                'refresh_token': self._api_keys['refresh_token'],
                'client_id': self._api_keys['client_id'],
                'client_secret': self._api_keys['client_secret'],
                'redirect_url': self._api_keys['redirect_url'],
                'grant_type': "refresh_token"
            })
            if rsp.ok:
                self.update_tokens(rsp.json())
                return {'Authorization': f"Zoho-oauthtoken {self._api_keys['access_token']}"}
            raise ZohoAuthRefreshFailure()
        raise ZohoInsufficentAuthKeys()

    def do_request(self, requestFunc, url, body=None, files=None):
        if self._api_keys['min_calls_remaining'] >= int(self._ratelimit['remaining']):
            if self._ratelimit['ResetAt'] > datetime.datetime.now():
                self.log("API call limit exceeded")
                raise ZohoAPICallsExceeded()

        now = datetime.datetime.now().timestamp()
        if now < self._ratelimit['NextCall']:
            self.log("pausing for internal API rate limit")
            sleep(self._ratelimit['NextCall'] - now)

        retries = self._api_keys['max_retries']
        while True:
            reqparams = {
                'headers': self.auth_header(),
                'files': files,
                'json': body,
            }
            rsp = requestFunc(url, **reqparams)
            self._ratelimit['NextCall'] = datetime.datetime.now().timestamp() + self._api_keys['intercall_delay']
            if rsp.status_code == 429 and retries:  # Too Many Requests
                retries -= 1
                sleeptime = int(rsp.headers.get('retry-after', self._api_keys['retry_backoff_seconds']))
                if sleeptime <= self._api_keys['max_retry_after']:
                    self.log("pausing before retry")
                    sleep(sleeptime)
                else:
                    raise ZohoAPIThrottled()
            elif not rsp.ok:
                err_params = {
                    'code': rsp.status_code,
                    'url': url,
                    'msg': f"Encountered #{rsp.status_code} error calling Zoho API",
                }
                if rsp.headers.get('content-type',"").startswith("application/json"):
                    d = rsp.json()
                    err_params.update({'zoho_code': d.get('code'), 'zoho_msg': d.get('message',"")})
                if rsp.status_code == 401:
                    self._api_keys['access_token'] = None
                    retries -= 1
                    if retries:
                        continue
                raise {
                    '400': ZohoBadRequest,
                    '401': ZohoUnauthorized,
                    '404': ZohoNotFound,
                    '405': ZohoMethodNotAllowed,
                }.get(str(rsp.status_code), ZohoException)(**err_params)
            else:
                if rsp.headers['content-type'].startswith("application/json"):
                    d = rsp.json()
                    if d.get('code') == 0:
                        self.update_rate_limit(rsp.headers)
                        return rsp
                    elif d.get('code') == "43" and retries:    # Throttled
                        retries -= 1
                        sleeptime = int(rsp.headers.get('retry-after', self._api_keys['retry_backoff_seconds']))
                        if sleeptime <= self._api_keys['max_retry_after']:
                            self.log("pausing before retry")
                            sleep(sleeptime)
                        else:
                            raise ZohoAPIThrottled()
                else:
                    return rsp

    def delete(self, urlFragment):
        url = f"{self._endpoint}/{urlFragment}?organization={self._org}"
        self.log(f"DELETE {url}")
        rsp = self.do_request(requests.delete, url)
        return rsp.ok

    def get(self, urlFragment, queryString):
        url = f"{self._endpoint}/{urlFragment}?organization={self._org}&{queryString}"
        self.log(f"GET {url}")
        rsp = self.do_request(requests.get, url)
        if rsp.headers['content-type'].startswith("application/json"):
            data = rsp.json()
            if data['code'] == 0:
                return data
            raise ZohoException(f"zoho returned {data['code']}: {data['message']}")
        return DottedDict({
            'content': rsp.content,
            'content_type': rsp.headers['content-type'],
        })

    def get_endpoint(self, region):
        # This MUST be overridden in subclasses for the APIs to work.
        # It's only here in the base class for testing
        return False

    def log(self, message, level=logging.DEBUG):
        self._logger.log(level, f"{self.__class__.__name__} (Org# {self._org}): {message}")

    def post(self, urlFragment, data=None, queryString="", files=None):
        url = f"{self._endpoint}/{urlFragment}?organization={self._org}&{queryString}"
        self.log(f"POST {url}")
        rsp = self.do_request(requests.post, url, data, files)
        if rsp.headers['content-type'].startswith("application/json"):
            data = rsp.json()
            if data['code'] == 0:
                return data
            raise ZohoException(f"zoho returned {data['code']}: {data['message']}")
        return {
            'content': rsp.content,
            'content_type': rsp.headers['content-type'],
        }

    def put(self, urlFragment, data, queryString):
        url = f"{self._endpoint}/{urlFragment}?organization={self._org}&{queryString}"
        self.log(f"PUT {url}")
        rsp = self.do_request(requests.put, url, data)
        if rsp.headers['content-type'].startswith("application/json"):
            data = rsp.json()
            if data['code'] == 0:
                return data
            raise ZohoException(f"zoho returned {data['code']}: {data['message']}")
        return {
            'content': rsp.content,
            'content_type': rsp.headers['content-type'],
        }

    def update_rate_limit(self, headers):
        for key in ['limit', 'reset', 'remaining']:
            if headers.get(f'x-rate-limit-{key}'):
                self._ratelimit[key] = headers[f'x-rate-limit-{key}']
        self._ratelimit['ResetAt'] = datetime.datetime.now() + datetime.timedelta(seconds=int(self._ratelimit.get('reset', 0)))

    def update_tokens(self, apiArgs):
        self._api_keys.update(apiArgs)
        self._api_keys['AccessExpiresAt'] = datetime.datetime.now().timestamp() + int(apiArgs.get('expires_in', 0)) - 10


class ZohoObjectBase:
    ID = property(lambda self: self._id)
    IsDeleted = property(lambda self: self._id is False)
    IsList = property(lambda self: isinstance(self._data, DottedList))
    IsLoaded = property(lambda self: self._data is not None)
    Number = property(lambda self: self._data[self._number_field] if isinstance(self._data, DottedDict) else None)

    def __init__(self, api, id=None, **searchParams):
        self._id = id
        self._api = api
        self._data = None
        self._nextpage = None
        if id or {k:v for k,v in searchParams.items() if isinstance(v, str)}:
            try:
                self._load(id=id, **searchParams)
            except ZohoNotFound as e:
                pass    # not found, but no need to raise this error.

    def __iter__(self):
        return self.Iter()

    def __getattr__(self, key):
        if self._data:
            return self._data[key]
        return None

    def __repr__(self):
        if self._id:
            return f"{self.__class__.__name__} #{self._id}"
        if isinstance(self._data, DottedList):
            return f"List of {self.__class__.__name__} objects"
        return f"New {self.__class__.__name__}"

    def __setattr__(self, key, value):
        if key.startswith('_'):
            super().__setattr__(key, value)
        else:
            if not self._data:
                self._data = DottedDict()
            self._data[key] = value

    def _load(self, page=None, **searchParams):
        data = self._api.get(self._url_fragment(), self._query_string(**searchParams))
        if data is None:
            self._data = None
        elif self._id and self._is_raw:
            self._data = DottedDict(data)
        else:
            if self._id:
                self._reload(data)
            else:
                if page:
                    self._data.extend(DottedList(data.get(self._plural)))
                else:
                    self._data = DottedList(data.get(self._plural))
                if data.get('page_context',{}).get('has_more_page'):
                    self._searchParams = searchParams
                    self._nextpage = data['page_context']['page'] + 1
                else:
                    self._nextpage = None

    @staticmethod
    def _passes_filter(obj, filterFunc=None, **filters):
        """ Utility function for filtering object lists
        """
        if filterFunc and not filterFunc(obj):
            return False
        for k,v in filters.items():
            if k in obj and obj[k] != v:
                return False
        return True

    def _query_string(self, **queryArgs):
        qs = "&".join([f"{k}={v}" for (k,v) in queryArgs.items() if v])
        if self._nextpage:
            qs += f"&page={self._nextpage}"
        return qs

    def _reload(self, data):
        self._orig = data.get(self._singular)
        self._data = DottedDict(self._orig)
        self._id = self._data.get(self._id_field)

    def _url_fragment(self, id=None, extraPath=[]):
        if self._id or id:
            return f"{self._type}/{id if id else self._id}/{'/'.join(extraPath)}"
        else:
            return self._type

    def get(self, key, default=None):
        if self._id:
            return self._data.get(key, default)
        return default

    def Create(self, **qParams):
        """ Create this object in Zoho

        :return: `self` as created by Zoho
        :raises ZohoInvalidOpError: if `self` isn't "new"
        """
        if not self._id:
            newData = self._api.post(self._url_fragment(), self._data.to_python(), self._query_string(**qParams))
            if newData:
                self._reload(newData)
            return self
        raise ZohoInvalidOpError("Create", self)

    def Delete(self):
        """ Delete this object from Zoho

        :return: `self`
        :raises ZohoInvalidOpError: if `self` isn't "single-object"
        """
        if self._id:
            if self._api.delete(self._url_fragment()):
                self._id = False
                self._data[self._id_field] = None
            return self
        raise ZohoInvalidOpError("Delete", self)

    def First(self, **kwargs):
        """ Get the first ZohoObject from the list.

        If kwargs are provided, they are used to filter what counts as "first."

        :return: a ZohoObject
        """
        if self._data == None:
            # We were a "new" object, but need to become a list-of
            self._load()
        if isinstance(self._data, DottedList) and len(self._data):
            if kwargs:
                rval = None
                for _ in self.Iter(raw=True, **kwargs): rval = _; break;
                if rval:
                    return self.__class__(self._api, rval[self._id_field])
            else:
                return self.__class__(self._api, self._data[0][self._id_field])
        return self.__class__(self._api)

    def GetRelated(self, targetType, key):
        if self._id and self._data:
            if isinstance(self._data.get(key), str):
                return targetType(self._data[key])
        raise ZohoInvalidOpError("GetRelated", self)

    def Iter(self, filterFunc=None, raw=False, **filter):
        """ Iterate over the list of ZohoObjects

        If called on a non-connected ("new") object, we get the list of ALL objects

        :param filterFunc: function which takes the item to test, returning True or False
        :param raw: return the raw list data?
        :param filter: fields/values to filter the list by
        :return: iterable of ZohoObjects (technically, a generator)
        """
        if self._id:
            return [self]
        if not self.IsLoaded:
            self._load()

        if self.IsLoaded:
            start = 0
            while start < len(self._data) or self._nextpage:
                if start >= len(self._data):
                    self._load(page=self._nextpage, **self._searchParams)

                item = self._data[start]
                if self._passes_filter(item, filterFunc=filterFunc, **filter):
                    if raw:
                        yield item
                    else:
                        yield self.__class__(self._api, id=item[self._id_field])
                start += 1
        else:
            return []

    def IterRelatedList(self, targetType, listKey, idField):
        if self._id and self._data:
            for item in self._data.get(listKey, []):
                yield targetType(item.get(idField))
        else:
            raise ZohoInvalidOpError("IterRelatedList", self)

    def MapRelatedList(self, targetType, listKey, idField):
        if self._id and self._data:
            for item in self._data.get(listKey, []):
                yield DottedDict({'meta': item.to_python(), 'object': targetType(item.get(idField))})
        else:
            raise ZohoInvalidOpError("MapRelatedList", self)

    def Update(self):
        if self._id and self._data:
            updated = diff(self._orig, self._data.to_python())
            if updated:
                # Maybe we don't need this; testing must ensue
                # for k in self._req_fields:
                #     if self._data.get(k):
                #         updated[k] = self._data[k]
                data = self._api.put(self._url_fragment(), updated, "")
                self._reload(data)
            return self
        raise ZohoInvalidOpError("Update", self)



# This file is part of pyZohoAPI, Copyright (C) Todd D. Esposito 2021.
# Distributed under the MIT License (see https://opensource.org/licenses/MIT).

from ..core import ZohoObjectBase
from .mixins import *

def ZohoObjectFactory(name,
    urlPath=None, responseKey=None, pluralResponseKey=None,
    idKey=None, numberKey=None, raw=False,
    mixins=[]):
    """ Factory function to create a Zoho Object type

    Unless overridden by the optional paremters, `name` is used to determine how
    Zoho API objects are accessed and parsed.

    * The URL used will be https:/{api-root}/{pluralized `name` -or- `urlPath`}
    * A list of objects is under the JSON key `pluralizedResponseKey` -or- `responseKey` pluralized -or- `name` pluralized
    * A single object of objects is under the JSON key `responseKey` -or- `name`
    * The object ID is under the JSON key `idKey` -or- `name` _id
    * The object Number is under the JSON key `numberKey` -or- `name` _number

    :param name: object-class name; the basis of the API URL path unless overridden.
    :param urlPath: sets th API URL path
    :param responseKey: sets the singular (and maybe plural) response key(s).
    :param pluralResponseKey: sets the plural response key.
    :param idKey: sets the id field key.
    :param numberKey: sets the number field key.
    :param raw: handle the response body as raw data rather than json for non-list responses
    :param mixins: list of object-type mixins. Defaults to [].
    :return: ZohoObject sub-class
    """
    return type(name, tuple([ZohoObjectBase] + mixins), {
        '__name__': name,
        '_type': urlPath if urlPath else f"{name.lower()}s",
        '_singular': responseKey if responseKey else name.lower(),
        '_plural': pluralResponseKey if pluralResponseKey else f"{responseKey}s" if responseKey else f"{name.lower()}s",
        '_id_field': idKey if idKey else f"{name.lower()}_id",
        '_number_field': numberKey if numberKey else f"{name.lower()}_number",
        '_is_raw': raw,
    })

Account = ZohoObjectFactory("ChartOfAccount", idKey="account_id")
Bill = ZohoObjectFactory("Bill", mixins=[HasCustomFields])
Brand = ZohoObjectFactory("Brand")
Bundle = ZohoObjectFactory("Bundle")
CompositeItem = ZohoObjectFactory("CompositeItem",
    responseKey="composite_item", idKey="composite_item_id",
    mixins=[HasActivate, HasBundle, HasCustomFields])
Contact = ZohoObjectFactory("Contact", mixins=[HasActivate, HasAddresses, HasCustomFields])
CreditNote = ZohoObjectFactory("CreditNode", mixins=[HasCustomFields])
Currency = ZohoObjectFactory("Currency", urlPath="settings/currencies",
    pluralResponseKey="currencies")
CustomerPayment = ZohoObjectFactory("CustomerPayment",
    idKey="payment_id", numberKey="payment_number", mixins=[HasCustomFields])
Document = ZohoObjectFactory("Document", raw=True)
Invoice = ZohoObjectFactory("Invoice", mixins=[HasCustomFields])
Item = ZohoObjectFactory("Item", mixins=[HasActivate, HasCustomFields, HasImage])
ItemAdjustment = ZohoObjectFactory("ItemAdjustment",
    urlPath="inventoryadjustments", responseKey="inventory_adjustment",
    idKey="inventory_adjustment_id")
ItemGroup = ZohoObjectFactory("ItemGroup", idKey="group_id",
    responseKey="group", mixins=[HasActivate])
Organization = ZohoObjectFactory("Organization")
Package = ZohoObjectFactory("Package", mixins=[HasCustomFields])
PriceList = ZohoObjectFactory("PriceList", urlPath="pricebooks",
    responseKey="pricebook", idKey="pricebook_id",
    mixins=[HasActivate])
PurchaseOrder = ZohoObjectFactory("PurchaseOrder", mixins=[HasCustomFields])
PurchaseReceive = ZohoObjectFactory("PurchaseReceive", mixins=[HasCustomFields])
RetainerInvoice = ZohoObjectFactory("RetainerInvoice", mixins=[HasDraft, HasVoid])
SalesOrder = ZohoObjectFactory("SalesOrder",
    mixins=[HasConfirm, HasCustomFields, HasVoid])
SalesPerson = ZohoObjectFactory("SalesPerson")
SalesReturn = ZohoObjectFactory("SalesReturn")
ShipmentOrder = ZohoObjectFactory("ShipmentOrder",
    responseKey="shipment_order", idKey="shipment_id",
    numberKey="shipment_number", mixins=[HasCustomFields, HasDelivered])
Tax = ZohoObjectFactory("Tax", urlPath="settings/taxes",
    pluralResponseKey="taxes")
TaxAuthority = ZohoObjectFactory("Tax", urlPath="settings/taxauthorities",
    responseKey="tax_authority", pluralResponseKey="tax_authorities",
    idKey="tax_authority_id")
TaxExemption = ZohoObjectFactory("Tax", urlPath="settings/taxexemptions",
    responseKey="tax_exemption", idKey="tax_exemption_id")
TaxGroup = ZohoObjectFactory("Tax", urlPath="settings/taxgroups",
    responseKey="tax_group", idKey="tax_group_id")
TransferOrder = ZohoObjectFactory("TransferOrder",
    responseKey="transfer_order",
    idKey="transfer_order_id", numberKey="transfer_order_number")
User = ZohoObjectFactory("User", mixins=[HasActivate, HasCustomFields])
Warehouse = ZohoObjectFactory("Warehouse", urlPath="settings/warehouses",
    mixins=[HasActivate])



# This file is part of pyZohoAPI, Copyright (C) Todd D. Esposito 2021.
# Distributed under the MIT License (see https://opensource.org/licenses/MIT).

from .core import ZohoAPIBase
from . import objecttypes

class ZohoInventory(ZohoAPIBase):
    _scope = "ZohoInventory.FullAccess.all"

    def get_endpoint(self, region):
        return f"https://inventory.zoho.{self._regionmap[region]}/api/v1"

    def Account(self, *args, **kwargs): return objecttypes.Account(self, *args, **kwargs)
    def Bill(self, *args, **kwargs): return objecttypes.Bill(self, *args, **kwargs)
    def Brand(self, *args, **kwargs): return objecttypes.Brand(self, *args, **kwargs)
    def Bundle(self, *args, **kwargs): return objecttypes.Bundle(self, *args, **kwargs)
    def CompositeItem(self, *args, **kwargs): return objecttypes.CompositeItem(self, *args, **kwargs)
    def Contact(self, *args, **kwargs): return objecttypes.Contact(self, *args, **kwargs)
    def CustomerPayment(self, *args, **kwargs): return objecttypes.CustomerPayment(self, *args, **kwargs)
    def Currency(self, *args, **kwargs): return objecttypes.Currency(self, *args, **kwargs)
    def Document(self, *args, **kwargs): return objecttypes.Document(self, *args, **kwargs)
    def Invoice(self, *args, **kwargs): return objecttypes.Invoice(self, *args, **kwargs)
    def Item(self, *args, **kwargs): return objecttypes.Item(self, *args, **kwargs)
    def ItemAdjustment(self, *args, **kwargs): return objecttypes.ItemAdjustment(self, *args, **kwargs)
    def ItemGroup(self, *args, **kwargs): return objecttypes.ItemGroup(self, *args, **kwargs)
    def Organization(self, *args, **kwargs): return objecttypes.Organization(self, *args, **kwargs)
    def Package(self, *args, **kwargs): return objecttypes.Package(self, *args, **kwargs)
    def PriceList(self, *args, **kwargs): return objecttypes.PriceList(self, *args, **kwargs)
    def PurchaseOrder(self, *args, **kwargs): return objecttypes.PurchaseOrder(self, *args, **kwargs)
    def PurchaseReceive(self, *args, **kwargs): return objecttypes.PurchaseReceive(self, *args, **kwargs)
    def RetainerInvoice(self, *args, **kwargs): return objecttypes.RetainerInvoice(self, *args, **kwargs)
    def SalesOrder(self, *args, **kwargs): return objecttypes.SalesOrder(self, *args, **kwargs)
    def SalesPerson(self, *args, **kwargs): return objecttypes.SalesPerson(self, *args, **kwargs)
    def SalesReturn(self, *args, **kwargs): return objecttypes.SalesReturn(self, *args, **kwargs)
    def ShipmentOrder(self, *args, **kwargs): return objecttypes.ShipmentOrder(self, *args, **kwargs)
    def Tax(self, *args, **kwargs): return objecttypes.Tax(self, *args, **kwargs)
    def TaxAuthority(self, *args, **kwargs): return objecttypes.TaxAuthority(self, *args, **kwargs)
    def TaxExemption(self, *args, **kwargs): return objecttypes.TaxExemption(self, *args, **kwargs)
    def TaxGroup(self, *args, **kwargs): return objecttypes.TaxGroup(self, *args, **kwargs)
    def TransferOrder(self, *args, **kwargs): return objecttypes.TransferOrder(self, *args, **kwargs)
    def User(self, *args, **kwargs): return objecttypes.User(self, *args, **kwargs)
    def Warehouse(self, *args, **kwargs): return objecttypes.Warehouse(self, *args, **kwargs)



# This file is part of pyZohoAPI, Copyright (C) Todd D. Esposito 2021.
# Distributed under the MIT License (see http://opensource.org/licenses/MIT).

from .core import ZohoAPIBase
from .exceptions import ZohoException
from . import objecttypes

class ZohoBooks(ZohoAPIBase):
    _scope = "ZohoBooks.FullAccess.all"

    def get_endpoint(self, region):
        return f"https://books.zoho.{self._regionmap[region]}/api/v3"

    def Account(self, *args, **kwargs): return objecttypes.Account(self, *args, **kwargs)
    def CompositeItem(self, *args, **kwargs): return objecttypes.CompositeItem(self, *args, **kwargs)
    def Contact(self, *args, **kwargs): return objecttypes.Contact(self, *args, **kwargs)
    def CustomerPayment(self, *args, **kwargs): return objecttypes.CustomerPayment(self, *args, **kwargs)
    def Document(self, *args, **kwargs): return objecttypes.Document(self, *args, **kwargs)
    def Invoice(self, *args, **kwargs): return objecttypes.Invoice(self, *args, **kwargs)
    def Item(self, *args, **kwargs): return objecttypes.Item(self, *args, **kwargs)
    def ItemGroup(self, *args, **kwargs): return objecttypes.ItemGroup(self, *args, **kwargs)
    def Organization(self, *args, **kwargs): return objecttypes.Organization(self, *args, **kwargs)
    def PurchaseOrder(self, *args, **kwargs): return objecttypes.PurchaseOrder(self, *args, **kwargs)
    def SalesOrder(self, *args, **kwargs): return objecttypes.SalesOrder(self, *args, **kwargs)
    def SalesPerson(self, *args, **kwargs): return objecttypes.SalesPerson(self, *args, **kwargs)
    def User(self, *args, **kwargs): return objecttypes.User(self, *args, **kwargs)



# This file is part of pyZohoAPI, Copyright (C) Todd D. Esposito 2021.
# Distributed under the MIT License (see https://opensource.org/licenses/MIT).

__version__ = '0.9.0'
__all__ = [
    "ZohoBooks",
    "ZohoInventory",
]

from .books import ZohoBooks
from .inventory import ZohoInventory



# -*- coding: utf-8 -*-

# The MIT License (MIT)
#
# Copyright (c) 2014 Carlos Escribano Rey
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

#==============================================================================#
# Modified slighly to include in pyZohoAPI
# We've removed py2 support, and added upstream-but-unmerged pullreq #14

import collections
import json
import re

from abc import ABCMeta, abstractmethod

class KeyOrAttributeError(KeyError, AttributeError): pass


SPLIT_REGEX = r"(?<!\\)(\.)"


def is_dotted_key(key):
    """Returns True if the key has any not-escaped dot inside"""
    return len(re.findall(SPLIT_REGEX, key)) > 0


def split_key(key, max_keys=0):
    parts = [x for x in re.split(SPLIT_REGEX, key) if x != "."]
    result = []
    while len(parts) > 0:
        if max_keys > 0 and len(result) == max_keys:
            break
        result.append(parts.pop(0))

    if len(parts) > 0:
        result.append(".".join(parts))
    return result


class DottedCollection(object, metaclass=ABCMeta):
    """Abstract Base Class for DottedDict and DottedDict"""

    @classmethod
    def factory(cls, initial=None):
        """Returns a DottedDict or a DottedList based on the type of the
        initial value, that must be a dict or a list. In other case the same
        original value will be returned.
        """
        if isinstance(initial, list):
            return DottedList(initial)
        elif isinstance(initial, dict):
            return DottedDict(initial)
        else:
            return initial

    @classmethod
    def load_json(cls, json_value):
        """Returns a DottedCollection from a JSON string"""
        return cls.factory(json.loads(json_value))

    @classmethod
    def _factory_by_index(cls, dotted_key):
        """Returns the proper DottedCollection that best suits the next key in
        the dotted_key string. First guesses the next key and then analyzes it.
        If the next key is numeric then returns a DottedList. In other case a
        DottedDict is returned.
        """
        if not isinstance(dotted_key, str):
            next_key = str(dotted_key)
        elif not is_dotted_key(dotted_key):
            next_key = dotted_key
        else:
            next_key, tmp = split_key(dotted_key, 1)

        return DottedCollection.factory([] if next_key.isdigit() else {})

    def __init__(self, initial):
        """Base constructor. If there are nested dicts or lists they are
        transformed into DottedCollection instances.
        """
        if not isinstance(initial, list) and not isinstance(initial, dict):
            raise ValueError('initial value must be a list or a dict')

        self._validate_initial(initial)

        self.store = initial

        if isinstance(self.store, list):
            data = enumerate(self.store)
        else:
            data = self.store.items()

        for key, value in data:
            try:
                self.store[key] = DottedCollection.factory(value)
            except ValueError:
                pass

    def _validate_initial(self, initial):
        """Validates data so no unescaped dotted key is present."""
        if isinstance(initial, list):
            for item in initial:
                self._validate_initial(item)
        elif isinstance(initial, dict):
            for key, item in initial.items():
                if is_dotted_key(key):
                    raise ValueError("{0} is not a valid key inside a "
                                     "DottedCollection!".format(key))
                self._validate_initial(item)

    def __len__(self):
        return len(self.store)

    def __iter__(self):
        return iter(self.store)

    def __repr__(self):
        return repr(self.store)

    def to_json(self):
        return json.dumps(self, cls=DottedJSONEncoder)

    @abstractmethod
    def __getitem__(self, name):
        pass

    @abstractmethod
    def __setitem__(self, name, value):
        pass

    @abstractmethod
    def __delitem__(self, name):
        pass

    @abstractmethod
    def to_python(self):
        pass


class DottedList(DottedCollection, collections.abc.MutableSequence):
    """A list with support for the dotted path syntax"""

    def __init__(self, initial=None):
        DottedCollection.__init__(
            self,
            [] if initial is None else list(initial)
        )

    def __getitem__(self, index):
        if isinstance(index, slice):
            return self.store[index]

        if isinstance(index, int) \
                or (isinstance(index, str) and index.isdigit()):
            return self.store[int(index)]

        elif isinstance(index, str) and is_dotted_key(index):
            my_index, alt_index = split_key(index, 1)
            target = self.store[int(my_index)]

            # required by the dotted path
            if not isinstance(target, DottedCollection):
                raise IndexError('cannot get "{0}" in "{1}" ({2})'.format(
                    alt_index,
                    my_index,
                    repr(target)
                ))

            return target[alt_index]

        else:
            raise IndexError('cannot get %s in %s' % (index, repr(self.store)))

    def __setitem__(self, index, value):
        if isinstance(index, int) \
                or (isinstance(index, str) and index.isdigit()):
            # If the index does not exist in the list but it's the same index
            # we would obtain by appending the value to the list we actually
            # append the value. (***)
            if int(index) not in self.store and int(index) == len(self.store):
                self.store.append(DottedCollection.factory(value))
            else:
                self.store[int(index)] = DottedCollection.factory(value)

        elif isinstance(index, str) and is_dotted_key(index):
            my_index, alt_index = split_key(index, 1)

            # (***)
            if int(my_index) not in self.store \
                    and int(my_index) == len(self.store):
                self.store.append(
                    DottedCollection._factory_by_index(alt_index))

            if not isinstance(self[int(my_index)], DottedCollection):
                raise IndexError('cannot set "%s" in "%s" (%s)' % (
                    alt_index, my_index, repr(self[int(my_index)])))

            self[int(my_index)][alt_index] = DottedCollection.factory(value)

        else:
            raise IndexError('cannot use %s as index in %s' % (
                index, repr(self.store)))

    def __delitem__(self, index):
        if isinstance(index, int) \
                or (isinstance(index, str) and index.isdigit()):
            del self.store[int(index)]

        elif isinstance(index, str) and is_dotted_key(index):
            my_index, alt_index = split_key(index, 1)
            target = self.store[int(my_index)]

            # required by the dotted path
            if not isinstance(target, DottedCollection):
                raise IndexError('cannot delete "%s" in "%s" (%s)' % (
                    alt_index, my_index, repr(target)))

            del target[alt_index]

        else:
            raise IndexError('cannot delete %s in %s' % (
                index, repr(self.store)))

    def to_python(self):
        """Returns a plain python list and converts to plain python objects all
        this object's descendants.
        """
        result = list(self)

        for index, value in enumerate(result):
            if isinstance(value, DottedCollection):
                result[index] = value.to_python()

        return result

    def insert(self, index, value):
        self.store.insert(index, value)


class DottedDict(DottedCollection, collections.abc.MutableMapping):
    """A dict with support for the dotted path syntax"""
    def __init__(self, initial=None):
        DottedCollection.__init__(
            self,
            {} if initial is None else dict(initial)
        )

    def __getitem__(self, k):
        key = self.__keytransform__(k)

        if not isinstance(k, str) or not is_dotted_key(key):
            try:
                return self.store[key]
            except KeyError as e:
                raise KeyOrAttributeError(e)

        my_key, alt_key = split_key(key, 1)
        target = self.store[my_key]

        # required by the dotted path
        if not isinstance(target, DottedCollection):
            raise KeyError('cannot get "{0}" in "{1}" ({2})'.format(
                alt_key,
                my_key,
                repr(target)
            ))

        return target[alt_key]

    def __setitem__(self, k, value):
        key = self.__keytransform__(k)

        if not isinstance(k, str):
            raise KeyError('DottedDict keys must be str or unicode')
        elif not is_dotted_key(key):
            self.store[key] = DottedCollection.factory(value)
        else:
            my_key, alt_key = split_key(key, 1)

            if my_key not in self.store:
                self.store[my_key] = DottedCollection._factory_by_index(alt_key)

            self.store[my_key][alt_key] = value

    def __delitem__(self, k):
        key = self.__keytransform__(k)

        if not isinstance(k, str) or not is_dotted_key(key):
            del self.store[key]

        else:
            my_key, alt_key = split_key(key, 1)
            target = self.store[my_key]

            if not isinstance(target, DottedCollection):
                raise KeyError('cannot delete "{0}" in "{1}" ({2})'.format(
                    alt_key,
                    my_key,
                    repr(target)
                ))

            del target[alt_key]

    def to_python(self):
        """Returns a plain python dict and converts to plain python objects all
        this object's descendants.
        """
        result = dict(self)

        for key, value in result.items():
            if isinstance(value, DottedCollection):
                result[key] = value.to_python()

        return result

    __getattr__ = __getitem__

    # self.store does not exist before __init__() initializes it

    def __setattr__(self, key, value):
        if key in self.__dict__ or key == 'store':
            object.__setattr__(self, key, value)
        else:
            self.__setitem__(key, value)

    def __delattr__(self, key):
        if key in self.__dict__ or key == 'store':
            object.__delattr__(self, key)
        else:
            self.__delitem__(key)

    def __contains__(self, k):
        key = self.__keytransform__(k)

        if not isinstance(k, str) or not is_dotted_key(key):
            return self.store.__contains__(key)

        my_key, alt_key = split_key(key, 1)
        target = self.store[my_key]

        if not isinstance(target, DottedCollection):
            return False

        return alt_key in target

    def __keytransform__(self, key):
        return key


#
# JSON stuff
#


class DottedJSONEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, DottedCollection):
            return obj.store
        else:
            return json.JSONEncoder.default(obj)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pyzohoapi',
 'pyzohoapi.core',
 'pyzohoapi.exceptions',
 'pyzohoapi.objecttypes',
 'pyzohoapi.objecttypes.mixins']

package_data = \
{'': ['*']}

install_requires = \
['requests>=2.25.1,<3.0.0']

setup_kwargs = {
    'name': 'pyzohoapi',
    'version': '0.9.0',
    'description': 'Pythonic access to Zoho APIs in the Finance Plus suite.',
    'long_description': '# pyZohoAPI (v0.9.0)\n **pyZohoAPI** provides Pythonic access to Zoho APIs in the Finance Plus suite:\n * **Books**\n * *Checkout*<sup>*</sup>\n * *Expense*<sup>*</sup>\n * **Inventory**\n * *Invoice*<sup>*</sup>\n * *Subscriptions*<sup>*</sup>\n\n<sup>*</sup> Support is planned, but not yet available.\n\n[![PyPI](https://img.shields.io/pypi/v/pyzohoapi)](https://pypi.org/project/pyzohoapi/)\n[![PyPI - Python Version](https://img.shields.io/pypi/pyversions/pyzohoapi)](https://pypi.org/project/pyzohoapi/)\n![License](https://img.shields.io/github/license/tdesposito/pyZohoAPI)\n[![Documentation Status](https://readthedocs.org/projects/pyzohoapi/badge/?version=latest)](https://pyzohoapi.readthedocs.io/en/latest/?badge=latest)\n\n## Installing pyZohoAPI\n<!-- start installation -->\n\nYou\'ll need at least **Python 3.6** to install pyZohoAPI.\n\n### Via PyPI\n```console\n$ python -m pip install pyzohoapi\n```\n\n### From Source\nWe use [Poetry](https://python-poetry.org/) for virtual environment and\ndependency management.\n```console\n$ git clone https://github.com/tdesposito/pyZohoAPI.git\n$ cd pyZohoAPI\n$ poetry install\n$ poetry build\n$ pip install dist/*.whl\n```\n<!-- end installation -->\n\n## Basic Usage\n\n<!-- start basic-usage -->\n```python\n>>> from pyzohoapi import ZohoInventory\n>>> api = ZohoInventory("{your-orginization-id}", "{your-region}",\n...   client_id="{your-client-id}",\n...   client_secret="{your-client-secret}",\n...   refresh_token="{your-refresh-token}"\n... )\n>>> contact = api.Contact(email="test@example.com").First()\n>>> contact.IsLoaded\nTrue\n>>> contact.first_name\n\'test\'\n>>> contact.first_name = "Changed"\n>>> contact.Update()\n```\n<!-- end basic-usage -->\n\nSee the [full documetation on ReadTheDocs](https://pyzohoapi.readthedocs.io/en/latest/)\n\n## Contributing\nPull Requests gladly considered!\n',
    'author': 'Todd Esposito',
    'author_email': 'todd@toddesposito.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/tdesposito/pyZohoAPI',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



# This file is part of pyZohoAPI, Copyright (C) Todd D. Esposito 2021.
# Distributed under the MIT License (see https://opensource.org/licenses/MIT).

from ...exceptions import *

from .CompositeItemOps import HasBundle
from .CustomFields import HasCustomFields

__all__ = [
    'HasActivate',
    'HasAddresses',
    'HasBundle',
    'HasConfirm', 'HasDraft', 'HasVoid', 'HasDelivered',
    'HasCustomFields',
    'HasImage',
]

class HasActivate:
    """Adds `Activate()` and `Deactivate()`"""
    def _do_operation(self, status, funcname):
        if self._id and self._data:
            try:
                self._api.post(self._url_fragment(extraPath=[status]))
                self.status = status
                return True
            except ZohoException as e:
                return False
        raise ZohoInvalidOpError(funcname, self)

    def Activate(self):
        return self._do_operation('active', "Activate")

    def Deactivate(self):
        return self._do_operation('inactive', "Deactivate")


class HasImage:
    """Adds `AddImage()`, `DeleteImage()` and `GetImage()`"""
    def AddImage(self, name, image, type="image/jpeg"):
        if self._id:
            file = {'image': (name, image, type)}
            return self._api.post(self._url_fragment(extraPath=['image']), files=file)
        raise ZohoInvalidOpError("AddImage", self)

    def DeleteImage(self):
        if self._id:
            return self._api.delete(self._url_fragment(extraPath=['image']))
        raise ZohoInvalidOpError("AddImage", self)

    def GetImage(self):
        if self._id and self._data:
            return self._api.get(self._url_fragment(extraPath=['image']), "")
        raise ZohoInvalidOpError("GetImage", self)


class _HasStatus:
    def _mark(self, status):
        if self._id:
            try:
                self._api.post(self._url_fragment(extraPath=['status', status]))
                return True
            except ZohoException as e:
                return False
        return None


class HasConfirm(_HasStatus):
    """Adds `Confirm()`"""
    def Confirm(self):
        if not self._id:
            raise ZohoInvalidOpError("Confirm", self)
        return self._mark('confirmed')


class HasDelivered(_HasStatus):
    """Adds `Delivered()`"""
    def Delivered(self):
        if not self._id:
            raise ZohoInvalidOpError("Delivered", self)
        return self._mark('delivered')


class HasDraft(_HasStatus):
    """Adds `Draft()`"""
    def Draft(self):
        if not self._id:
            raise ZohoInvalidOpError("Draft", self)
        return self._mark('draft')


class HasVoid(_HasStatus):
    """Adds `Void()`"""
    def Void(self):
        if not self._id:
            raise ZohoInvalidOpError("Void", self)
        return self._mark('void')


class _HasAspect:
    def _updateAspect(self, aspect, aspect_id, data):
        data = self._api.put(self._url_fragment(extraPath=[aspect, aspect_id]), data, "")
        if data['code'] == 0:
            return True
        return False


class HasAddresses(_HasAspect):
    """Adds `UpdateBilling()` and `UpdateShipping()`"""
    def UpdateBilling(self):
        if self._id and self._data:
            self._updateAspect('address', self.billing_address.address_id, self.billing_address.to_python())
            return self
        raise ZohoInvalidOpError("UpdateBilling", self)
    def UpdateShipping(self):
        if self._id and self._data:
            self._updateAspect('address', self.shipping_address.address_id, self.shipping_address.to_python())
            return self
        raise ZohoInvalidOpError("UpdateShipping", self)



# This file is part of pyZohoAPI, Copyright (C) Todd D. Esposito 2021.
# Distributed under the MIT License (see https://opensource.org/licenses/MIT).
import datetime

from ...core.collection import DottedList

from ...exceptions import *

class HasCustomFields:
    def GetCustomField(self, key, listKey="custom_fields", default=None):
        if self._id:
            if self._data:
                for cf in self._data.get(listKey, []):
                    if key in [cf['label'], cf['placeholder']]:
                        return cf['value']
            return default
        raise ZohoInvalidOpError("GetCustomField", self)

    def SetCustomField(self, key, value, listKey="custom_fields", altkey=None):
        """ Sets the value of an existing Custom Field

        :param key: Custom Field `placeholder` or `label` (string)
        :param value: new value for the Custom Field.
        :param listKey: key to the Custom Fields list. Defaults to "custom_fields".
        :param altkey: "alternate" key for adding a custom field.
        :return: `self`
        :raises ZohoInvalidOpError: If `self` isn't an existing object.
        """
        if self._id:
            if self._data:
                for cf in self._data.get(listKey, []):
                    if key in [cf['label'], cf['placeholder']]:
                        cf['value'] = value
                        return self
                # We didn't find the custom field, so we need to add it.
                self._data.setdefault(listKey, DottedList()).append({
                    'placeholder': key if key.startswith("cf_") else altkey,
                    'label': altkey if key.startswith("cf_") else key,
                    'value': value,
                })
            return self
        raise ZohoInvalidOpError("SetCustomField", self)



# This file is part of pyZohoAPI, Copyright (C) Todd D. Esposito 2021.
# Distributed under the MIT License (see https://opensource.org/licenses/MIT).

def diff(orig, new):
    """ Calculates the difference between two dictionaries.

    Any key with a child list or dict which is, itself, changed will be
    considered changed.

    :param orig: the original (unmodified) dictionary
    :param new: the modified dictionary
    :return: a dictionary contianing only those keys which were changed.
    """
    updated = {}
    for k, v in new.items():
        if k not in orig:
            updated[k] = v
        elif isinstance(v, list):
            if len(v) != len(orig[k]):
                updated[k] = v
            else:
                has_change = False
                for i in range(len(v)):
                    if isinstance(v[i], dict) and diff(orig[k][i], v[i]):
                        has_change = True
                        break
                    elif v[i] != orig[k][i]:
                        has_change = True
                        break
                if has_change:
                    # the update needs to contain the ENTIRE new list, so
                    # Zoho doesn't zap non-updated elements
                    updated[k] = v
        elif isinstance(v, dict):
            if diff(v, orig[k]):
                # the update needs to contain the ENTIRE new dict, so
                # Zoho doesn't zap non-updated values
                updated[k] = v
        elif v != orig[k]:
            updated[k] = v
    return updated



# This file is part of pyZohoAPI, Copyright (C) Todd D. Esposito 2021.
# Distributed under the MIT License (see https://opensource.org/licenses/MIT).

class ZohoException(Exception):
    def __init__(self, msg="Unspecified Error", **kwargs):
        super().__init__(msg)

#---------------------------------------------------------------------------
# Exceptions raised by API objects

class ZohoAuthRefreshFailure(ZohoException):
    def __init__(self):
        super().__init__("Unable to refresh access token")


class ZohoInsufficentAuthKeys(ZohoException):
    def __init__(self):
        super().__init__("Unable to generate Authorization header")


class ZohoUnknownRegionException(ZohoException):
    def __init__(self, region):
        super().__init__(f'Unknown Region "{region}"')


#---------------------------------------------------------------------------
# Exceptions raised by Zoho objects

class ZohoAPICallsExceeded(ZohoException):
    def __init__(self):
        super().__init__("No API Calls remaining")


class ZohoAPIThrottled(ZohoException):
    def __init__(self):
        super().__init__("API Throttled")



class ZohoBadRequest(ZohoException):
    def __init__(self, url, zoho_code="-no code-", zoho_msg="-no message-", **kwargs):
        super().__init__(f"Bad Request (HTTP-400)\n{zoho_msg} (Zoho-{zoho_code})")


class ZohoUnauthorized(ZohoException):
    def __init__(self, **kwargs):
        super().__init__(f"Unauthorized - Invalid Access Token (HTTP=401)")


class ZohoNotFound(ZohoException):
    def __init__(self, url, **kwargs):
        super().__init__(f"Not Found: '{url}' (HTTP-404)")


class ZohoMethodNotAllowed(ZohoException):
    def __init__(self, url, **kwargs):
        super().__init__(f"Method not allowed on '{url}' (HTTP-405)")


class ZohoInvalidOpError(ZohoException):
    def __init__(self, op, target):
        if target.ID is False:
            super().__init__(f"Can't call {op} on a deleted {target.__class__.__name__}")
        elif target.IsList:
            super().__init__(f"Can't call {op} on a list-of {target.__class__.__name__}")
        else:
            # TODO: revise this message, provided we find a use for it
            super().__init__(f"Can't call {op} on an object of type {target.__class__.__name__}")



# This file is part of pyZohoAPI, Copyright (C) Todd D. Esposito 2021.
# Distributed under the MIT License (see https://opensource.org/licenses/MIT).
import datetime

from ...exceptions import *

class HasBundle:
    def MakeBundle(self, qty, bundle_date=None, purchase_account_id=None, warehouse_id=None):
        """ Create a bundle for a CompositeItem

        :param qty: Number of bundles to make.
        :param bundle_date: Date of bundling, in "YYYY-MM-DD" format. Defaults to today.
        :param purchase_account_id: ID of the purchase account. Defaults to None.
        :param warehouse_id: ID of the warehouse to make the bundle in. Defaults to None.
        :return: Bundle ZohoObject
        :raises ZohoInvalidOpError: if `self` is new, a list, or not loaded.
        """
        if self.IsLoaded and self.ID:
            bundle = self._api.Bundle()
            bundle.date = bundle_date if bundle_date else datetime.datetime.today().strftime("%Y-%m-%d")
            bundle.composite_item_id = self.ID
            bundle.composite_item_name = self.name
            if self.get('sku'):
                bundle.composite_item_sku = self.get('sku')
            bundle.quantity_to_bundle = qty
            bundle.is_completed = True
            bundle.line_items = []
            for line in self.mapped_items:
                bundle_line = {
                    'item_id': line['item_id'],
                    'name': line['name'],
                    'unit': line['unit'],
                    'rate': line['purchase_rate'],
                    'quantity_consumed': line['quantity'],
                }
                if purchase_account_id:
                    bundle_line['account_id'] = purchase_account_id
                if warehouse_id:
                    bundle_line['warehouse_id'] = warehouse_id
                bundle.line_items.append(bundle_line)

            return bundle.Create()
        raise ZohoInvalidOpError('MakeBundle', self)
